//	--------------------------------------------------------------------------
//	--------------------------------------------------------------------------
//	SQUARE SCREW THREAD
//
//		Based on the 'Square Screw Thread' program from Leendert Ammeral
//		on his book 'Interactive 3d computer graphics' - ISBN 0-471-92014-2
//
//		louis bellotto - louisbel@free.fr - janvier 2004
//	--------------------------------------------------------------------------
//	--------------------------------------------------------------------------
#version unofficial MegaPov 1.0;




//	--------------------------------------------------------------------------
//	--- MACRO : MAKE OBJECT SQUARE SCREW THREAD ------------------------------
//	--------------------------------------------------------------------------
//	This macro generate an object directly in the scene.
//
//	Parameters :
//		- MajorDiameter : major diameter of the screw
//		- MinorDiameter : minor diameter of the screw
//		- ScrewLen : len of the screw
//		- StepAngle : angle. in fact precision.
//
//	Note : 	for square screw thread, pitch = MajorDiameter - MinorDiameter
//				so, you can pass 'MajorDiameter-pitch' for MinorDiameter parameter
//	--------------------------------------------------------------------------
#macro MakeObjectSquareScrew (MajorDiameter, MinorDiameter, ScrewLen, StepAngle)

	#macro addOnePoint(a)
		<p[a].x, p[a].y>,
	#end
	
	#macro addLastPoint(a)
		<p[a].x, p[a].y>
	#end
	
	#macro addRangePoints(a,b)
		#local i=0;
		#if (b>=a)
			#set i=a;
			#while (i<=b)
				<p[i].x, p[i].y>,
				#set i=i+1;
			#end			
		#else
			#set i=a;
			#while (i>=b)
				<p[i].x, p[i].y>,
				#set i=i-1;
			#end
		#end
	#end

	#macro writeTriangle(a,b,c)
		triangle { <p[a].x, p[a].y, p[a].z>, <p[b].x, p[b].y, p[b].z>, <p[c].x, p[c].y, p[c].z> }
	#end
	
	#local k = 180/StepAngle;
	#local pitch = MajorDiameter-MinorDiameter;
	#local h = pitch/2;
	#local n = 2*k;
	#local deltaL = pitch/n;
	#local step = int((ScrewLen/deltaL) + 0.50);
	#local N = step+1;
	#local N2 = 2*N;
	#local N3 = 3*N;
	#local N4 = 4*N;
	#local L = step*deltaL;
	#local theta = pi/k;
	#local R = MajorDiameter/2;
	#local r = MinorDiameter/2;
	#local A = 4*N-1;
	#local B = A+k-1;
	#local C = B+k-1;
	#local D = C+k-1;
	
	#local nbPoints = D+k+1;
	#local p = array[nbPoints];
	
	#local index=0;
	#local phi = 0.00;
	
	#debug concat("\n\nComputing ",str(nbPoints-1,0,0)," vertices.\n\n")

	// --- outer helix on upper flank -------------------------------------------
	#set index = 1;
	#while (index<=N)
		#set phi = (index-1)*theta;
		#local p[index] = <R*cos(phi), R*sin(phi), (index-1)*deltaL>;
		#set index=index+1;
	#end
	
	// --- inner helix on upper flank -------------------------------------------
	#while (index<=N2)
		#local p[index] = <p[index-N].x*r/R, p[index-N].y*r/R, p[index-N].z>;
		#set index=index+1;
	#end
	
	// --- outer helix on lower flank -------------------------------------------
	#while (index<=N3)
		#local p[index] = <-p[index-N2].x, -p[index-N2].y, p[index-N2].z>;
		#set index=index+1;
	#end
	
	// --- inner helix on lower flank -------------------------------------------
	#while (index<=N4)
		#local p[index] = <p[index-N].x*r/R, p[index-N].y*r/R, p[index-N].z>;
		#set index=index+1;
	#end
	
	// -- outer half circle in bottom plane -------------------------------------
	#while (index<=A+k)
		#set phi=(index-A-1)*theta;
		#local p[index] = <R*cos(phi), R*sin(phi), 0.00>;
		#set index=index+1;
	#end
	
	// --- inner half circle in bottom plane ------------------------------------
	#while (index<=B+k)
		#local p[index] = <-p[index-B+A].x*r/R, -p[index-B+A].y*r/R, 0.00>;
		#set index=index+1;
	#end
	
	//	--------------------------------------------------------------------------
	#declare off_set = step*theta;
	#while (off_set>2*pi)
		#set off_set = off_set-2*pi;
	#end
	
	// --- outer half circle plane ----------------------------------------------
	#while (index<=C+k)
		#set phi = off_set+(index-C-1)*theta;
		#local p[index] = <R*cos(phi), R*sin(phi), L>;
		#set index=index+1;
	#end
	
	// --- inner half circle in top plane ---------------------------------------
	#while (index<=D+k)
		#local p[index] = <-p[index-D+C].x*r/R, -p[index-D+C].y*r/R, L>;
		#set index=index+1;
	#end
	
	//	--- building faces --------------------------------------------------------
	#debug "\n\nAssembling faces.\n\n"


	// --- z = 0 ----------------------------------------------------------------
	#local polyPointsCount = 7 + abs((A+k)-(A+2)) + abs((B+k)-(B+2)) ;
	#local lowerPolygon = polygon {
		polyPointsCount,
		addOnePoint(N2+1)
		addRangePoints(A+k,A+2)
		addOnePoint(1)
		addOnePoint(N+1)
		addRangePoints(B+k,B+2)
		addOnePoint(N3+1)
		addLastPoint(N2+1)
		translate p[N2+1].z*z		// alway zero !
		}

	// --- z = L ----------------------------------------------------------------
	#local polyPointsCount = 7 + abs((C+2)-(C+k)) + abs((D+2)-(D+k));
	#local upperPolygon = polygon {
		polyPointsCount,
		addOnePoint(N)
		addRangePoints(C+2, C+k)
		addOnePoint(N3)
		addOnePoint(N4)
		addRangePoints(D+2, D+k)
		addOnePoint(N2)
		addLastPoint(N)
		translate p[N].z*z
		}

	#local screwSurface = mesh {
		// --- first 180 degrees at the bootom --------------------------------------
	
		// --- outer surface --------------------------------------------------------
		writeTriangle(1,A+2,2)
		#set index=2;
		#while (index<k)
			writeTriangle(A+index, A+index+1, index+1)
			writeTriangle(A+index, index+1, index)
			#set index=index+1;
		#end
		writeTriangle(A+k, N2+1, k+1)
		writeTriangle(A+k,  k+1, k)
		
		// --- inner surface --------------------------------------------------------
		writeTriangle(B+2, N3+2, N3+1)
		#set index=2;
		#while (index<k)
			writeTriangle(B+index+1, N3+index+1, N3+index)
			writeTriangle(B+index+1,N3+index, B+index)
			#set index=index+1;
		#end
		writeTriangle(N+1, N3+k+1, N3+k)
		writeTriangle(N+1, N3+k B+k)
	
		// --- final 180 degrees at the top -----------------------------------------
	
		// --- outer surface --------------------------------------------------------
		writeTriangle(N3, C+k, N3-1)
		#set index=2;
		#while (index<k)
			writeTriangle(C+k-index+2, C+k-index+1, N3-index)
			writeTriangle(C+k-index+2, N3-index, N3-index+1)
		#set index=index+1;
		#end
		writeTriangle(C+2, N, N3-k)
		writeTriangle(C+2, N3-k, N3-k+1)
		
		// --- inner surface --------------------------------------------------------
		// p3e(D+k, N2-1, N2)
		writeTriangle(D+k, N2-1, N2)
		#set index=2;
		#while (index<k)
			writeTriangle(D+k-index+2, D+k-index+1, N2-index)
			writeTriangle(D+k-index+2, N2-index, N2-index+1)
			#set index=index+1;
		#end
		writeTriangle(D+2, N4, N2-k)
		writeTriangle(D+2, N2-k, N2-k+1)
	
		// --- middle part ----------------------------------------------------------
	
		// --- outer surface --------------------------------------------------------
		#set index=1;
		#while (index<N-k)
			writeTriangle(N2+index, N2+index+1, k+index+1)
			writeTriangle(N2+index, k+index+1, k+index)
			#set index=index+1;
		#end
		
		// --- inner surface --------------------------------------------------------
		#set index=1;
		#while (index<N-k)
			writeTriangle(N+index, N+index+1, N3+k+index+1)
			writeTriangle(N+index, N3+k+index+1, N3+k+index)
			#set index=index+1;
		#end
		
		//	--- upper flank ---------------------------------------------------------- 
		#set index=1;
		#while (index<N)
			writeTriangle(index, N+index+1, N+index)
			writeTriangle(N+index+1, index, index+1)
			#set index=index+1;
		#end
		
		//	--- lower flank ---------------------------------------------------------- 
		#set index=1;
		#while (index<N)
			writeTriangle(N3-index+1, N4-index, N4-index+1)
			writeTriangle(N4-index, N3-index+1, N3-index)
			#set index=index+1;
		#end
		
		inside_vector <0.00, 0.00, ScrewLen> // to make mesh solid, but seems not works well...
		} // end of mesh
	
	union {
		object { upperPolygon }
		object { lowerPolygon }
		object { screwSurface }
		}
#end

//	--------------------------------------------------------------------------
//	--- MACRO : MAKE FILE SQUARE SCREW THREAD TO FILE ------------------------
//	--------------------------------------------------------------------------
//	Same as above but square screw object is written to the file.
//	So, you can include generated file directly to your scene without
//	time consuming from calculation.
//
//	In this case, four globals variables are declared. you can use them in
// your scene :
//		#declare MajorDiameter : the Major DIAMETER of the screw.
//		#declare MinorDiameter : the minor DIAMETER of the screw.
//		#declare ScrewLen : the len of the screw along the z axis.
//		#declare StepAngle : step angle used for the construction, precision in fact.
//	--------------------------------------------------------------------------
#macro MakeIncludeSquareScrew (FileName, MajorDiameter, MinorDiameter, ScrewLen, StepAngle)

	#macro writeOnePoint(a)
		#write (theFileRef,"	<",p[a].x,", ",p[a].y,">,\n")
	#end
	
	#macro writeLastPoint(a)
		#write (theFileRef,"	<",p[a].x,", ",p[a].y,">\n")
	#end
	
	#macro writeRangePoints(a,b)
		#local i=0;
		#if (b>=a)
			#set i=a;
			#while (i<=b)
				#write (theFileRef,"	<",p[i].x,", ",p[i].y,">,\n")
				#set i=i+1;
			#end			
		#else
			#set i=a;
			#while (i>=b)
				#write (theFileRef,"	<",p[i].x,", ",p[i].y,">,\n")
				#set i=i-1;
			#end
		#end
	#end

	#macro writeTriangle(a,b,c)
		#write (theFileRef,"	triangle { <",p[a].x,", ",p[a].y,", ",p[a].z,">, <",p[b].x,", ",p[b].y,", ",p[b].z,">, <",p[c].x,", ",p[c].y,", ",p[c].z"> }\n")
	#end
	
	#local k = 180/StepAngle;
	#local pitch = MajorDiameter-MinorDiameter;
	#local h = pitch/2;
	#local n = 2*k;
	#local deltaL = pitch/n;
	#local step = int((ScrewLen/deltaL) + 0.50);
	#local N = step+1;
	#local N2 = 2*N;
	#local N3 = 3*N;
	#local N4 = 4*N;
	#local L = step*deltaL;
	#local theta = pi/k;
	#local R = MajorDiameter/2;
	#local r = MinorDiameter/2;
	#local A = 4*N-1;
	#local B = A+k-1;
	#local C = B+k-1;
	#local D = C+k-1;
	
	#local decimalPrecision = 2;
	
	#local nbPoints = D+k+1;
	#local p = array[nbPoints];
	
	#local index=0;
	#local phi = 0.00;
	
	#debug concat("\n\nComputing ",str(nbPoints-1,0,0)," vertices.\n\n")

	// --- outer helix on upper flank -------------------------------------------
	#set index = 1;
	#while (index<=N)
		#set phi = (index-1)*theta;
		#local p[index] = <R*cos(phi), R*sin(phi), (index-1)*deltaL>;
		#set index=index+1;
	#end
	
	// --- inner helix on upper flank -------------------------------------------
	#while (index<=N2)
		#local p[index] = <p[index-N].x*r/R, p[index-N].y*r/R, p[index-N].z>;
		#set index=index+1;
	#end
	
	// --- outer helix on lower flank -------------------------------------------
	#while (index<=N3)
		#local p[index] = <-p[index-N2].x, -p[index-N2].y, p[index-N2].z>;
		#set index=index+1;
	#end
	
	// --- inner helix on lower flank -------------------------------------------
	#while (index<=N4)
		#local p[index] = <p[index-N].x*r/R, p[index-N].y*r/R, p[index-N].z>;
		#set index=index+1;
	#end
	
	// -- outer half circle in bottom plane -------------------------------------
	#while (index<=A+k)
		#set phi=(index-A-1)*theta;
		#local p[index] = <R*cos(phi), R*sin(phi), 0.00>;
		#set index=index+1;
	#end
	
	// --- inner half circle in bottom plane ------------------------------------
	#while (index<=B+k)
		#local p[index] = <-p[index-B+A].x*r/R, -p[index-B+A].y*r/R, 0.00>;
		#set index=index+1;
	#end
	
	//	--------------------------------------------------------------------------
	#declare off_set = step*theta;
	#while (off_set>2*pi)
		#set off_set = off_set-2*pi;
	#end
	
	// --- outer half circle plane ----------------------------------------------
	#while (index<=C+k)
		#set phi = off_set+(index-C-1)*theta;
		#local p[index] = <R*cos(phi), R*sin(phi), L>;
		#set index=index+1;
	#end
	
	// --- inner half circle in top plane ---------------------------------------
	#while (index<=D+k)
		#local p[index] = <-p[index-D+C].x*r/R, -p[index-D+C].y*r/R, L>;
		#set index=index+1;
	#end
	
	//	--- building faces --------------------------------------------------------
	#debug "\n[Openning file]\n"
	#fopen theFileRef FileName write
	
	#debug "\nWritting to file..."
	#write (theFileRef,"// -----------------------------------------------------------------------------\n")
	#write (theFileRef,"// square screw thread\n")
	#write (theFileRef,"//\n")
	#write (theFileRef,"// macro by louis bellotto - comments are welcome - louisbel@noos.fr\n")
	#write (theFileRef,"// -----------------------------------------------------------------------------\n\n")
	#write (theFileRef,"#declare MajorDiameter = ",str(MajorDiameter,0,decimalPrecision),";\n")
	#write (theFileRef,"#declare MinorDiameter = ",str(MinorDiameter,0,decimalPrecision),";\n")
	#write (theFileRef,"#declare ScrewLen = ",str(ScrewLen,0,decimalPrecision),";\n")
	#write (theFileRef,"#declare StepAngle = ",str(StepAngle,0,decimalPrecision),";\n\n\n")

	// --- z = 0 ----------------------------------------------------------------
	#local polyPointsCount = 7 + abs((A+k)-(A+2)) + abs((B+k)-(B+2));
	
	#write (theFileRef,"#declare lowerPolygon = polygon {\n")
	#write (theFileRef,"	",polyPointsCount,",\n")
	
		writeOnePoint(N2+1)
		writeRangePoints(A+k,A+2)
		writeOnePoint(1)
		writeOnePoint(N+1)
		writeRangePoints(B+k,B+2)
		writeOnePoint(N3+1)
		writeLastPoint(N2+1)
	
	#write (theFileRef,"	translate ",p[N2+1].z,"*z		// alway zero !\n")
	#write (theFileRef,"	}\n\n")
	
	// --- z = L ----------------------------------------------------------------
	#local polyPointsCount = 7 + abs((C+2)-(C+k)) + abs((D+2)-(D+k));

	#write (theFileRef,"#declare upperPolygon = polygon {\n")
	#write (theFileRef,"	",polyPointsCount,",\n")

		writeOnePoint(N)
		writeRangePoints(C+2, C+k)
		writeOnePoint(N3)
		writeOnePoint(N4)
		writeRangePoints(D+2, D+k)
		writeOnePoint(N2)
		writeLastPoint(N)

	#write (theFileRef,"	translate ",p[N].z,"*z\n")
	#write (theFileRef,"	}\n\n")

	#write (theFileRef,"#declare screwSurface = mesh {\n")
		// --- first 180 degrees at the bootom --------------------------------------
	
		// --- outer surface --------------------------------------------------------
		writeTriangle(1,A+2,2)
		#set index=2;
		#while (index<k)
			writeTriangle(A+index, A+index+1, index+1)
			writeTriangle(A+index, index+1, index)
			#set index=index+1;
		#end
		writeTriangle(A+k, N2+1, k+1)
		writeTriangle(A+k,  k+1, k)
		
		// --- inner surface --------------------------------------------------------
		writeTriangle(B+2, N3+2, N3+1)
		#set index=2;
		#while (index<k)
			writeTriangle(B+index+1, N3+index+1, N3+index)
			writeTriangle(B+index+1,N3+index, B+index)
			#set index=index+1;
		#end
		writeTriangle(N+1, N3+k+1, N3+k)
		writeTriangle(N+1, N3+k B+k)
	
		// --- final 180 degrees at the top -----------------------------------------
	
		// --- outer surface --------------------------------------------------------
		writeTriangle(N3, C+k, N3-1)
		#set index=2;
		#while (index<k)
			writeTriangle(C+k-index+2, C+k-index+1, N3-index)
			writeTriangle(C+k-index+2, N3-index, N3-index+1)
		#set index=index+1;
		#end
		writeTriangle(C+2, N, N3-k)
		writeTriangle(C+2, N3-k, N3-k+1)
		
		// --- inner surface --------------------------------------------------------
		// p3e(D+k, N2-1, N2)
		writeTriangle(D+k, N2-1, N2)
		#set index=2;
		#while (index<k)
			writeTriangle(D+k-index+2, D+k-index+1, N2-index)
			writeTriangle(D+k-index+2, N2-index, N2-index+1)
			#set index=index+1;
		#end
		writeTriangle(D+2, N4, N2-k)
		writeTriangle(D+2, N2-k, N2-k+1)
	
		// --- middle part ----------------------------------------------------------
	
		// --- outer surface --------------------------------------------------------
		#set index=1;
		#while (index<N-k)
			writeTriangle(N2+index, N2+index+1, k+index+1)
			writeTriangle(N2+index, k+index+1, k+index)
			#set index=index+1;
		#end
		
		// --- inner surface --------------------------------------------------------
		#set index=1;
		#while (index<N-k)
			writeTriangle(N+index, N+index+1, N3+k+index+1)
			writeTriangle(N+index, N3+k+index+1, N3+k+index)
			#set index=index+1;
		#end
		
		//	--- upper flank ---------------------------------------------------------- 
		#set index=1;
		#while (index<N)
			writeTriangle(index, N+index+1, N+index)
			writeTriangle(N+index+1, index, index+1)
			#set index=index+1;
		#end
		
		//	--- lower flank ---------------------------------------------------------- 
		#set index=1;
		#while (index<N)
			writeTriangle(N3-index+1, N4-index, N4-index+1)
			writeTriangle(N4-index, N3-index+1, N3-index)
			#set index=index+1;
		#end
		
	#write (theFileRef,"	inside_vector <0.00, 0.00, ",ScrewLen*0.50,">\n") 
	#write (theFileRef,"	} // end of mesh\n\n\n")
	#write (theFileRef,"#declare theSquareScrew = union {\n")
	#write (theFileRef,"	object { upperPolygon }\n")
	#write (theFileRef,"	object { lowerPolygon }\n")
	#write (theFileRef,"	object { screwSurface }\n")
	#write (theFileRef,"	}\n\n\n")
		
	#local stampString = concat(date("%d"),"/",date("%m"),"/",date("%Y")," at ",date("%H:%M:%S"))
	#write (theFileRef,"// --- genered by the 'MakeIncludeSquareScrew' macro - ",stampString,"\n")
		
	#debug "\n[Closing file]\n"
	#fclose theFileRef

#end
//	--------------------------------------------------------------------------
//	--- END OF MACRO ---------------------------------------------------------
//	--------------------------------------------------------------------------




